# type: ignore
import anyio
from anyio import get_cancelled_exc_class

# this one is fine to also run with TRIO


async def foo():
    ...


async def foo_anyio():
    try:
        ...
    except anyio.get_cancelled_exc_class():
        await foo()  # error: 8, Statement("anyio.get_cancelled_exc_class()", lineno-1)
    except:
        await foo()  # safe

    try:
        ...
    except anyio.get_cancelled_exc_class():
        await foo()  # error: 8, Statement("anyio.get_cancelled_exc_class()", lineno-1)
    except BaseException:
        await foo()  # safe

    try:
        ...
    except get_cancelled_exc_class():
        await foo()  # error: 8, Statement("get_cancelled_exc_class()", lineno-1)
    except:
        await foo()  # safe

    try:
        ...
    except anyio.get_cancelled_exc_class():
        await foo()  # error: 8, Statement("anyio.get_cancelled_exc_class()", lineno-1)
    except anyio.get_cancelled_exc_class():
        await foo()

    # these are all type errors, and will not be matched by critical_except
    try:
        ...
    except anyio.get_cancelled_exc_class:
        await foo()  # not handled
    except anyio.get_cancelled_exc_class(...):
        await foo()  # not handled
    except get_cancelled_exc_class:
        await foo()  # not handled
    except:
        await foo()  # error: 8, Statement("bare except", lineno-1)

    try:
        ...
    except anyio.foo():
        await foo()
    except ValueError:
        await foo()  # safe
    except anyio.Cancelled:
        await foo()  # not an error
    except:
        await foo()  # error: 8, Statement("bare except", lineno-1)


async def foo_anyio_shielded():
    try:
        ...
    except anyio.get_cancelled_exc_class():
        with anyio.CancelScope(deadline=30, shield=True):
            await foo()  # safe
    except BaseException:
        await foo()  # safe



# mypy: disable-error-code="arg-type,attr-defined"
from contextlib import asynccontextmanager

import anyio
import trio

# NOANYIO - requires no substitution check


@asynccontextmanager
async def foo():
    with trio.open_nursery() as bar:
        bar.start_soon(trio.run_process)  # TRIO113: 8
    async with trio.open_nursery() as bar:
        bar.start_soon(trio.run_process)  # TRIO113: 8

    async with anyio.create_task_group() as bar_tg:
        bar_tg.start_soon(anyio.run_process)  # TRIO113: 8

    boo: trio.Nursery = ...  # type: ignore
    boo.start_soon(trio.run_process)  # TRIO113: 4

    boo_anyio: anyio.TaskGroup = ...  # type: ignore
    boo_anyio.start_soon(anyio.run_process)  # TRIO113: 4

    yield



# NOANYIO
import trio

try:
    ...
except trio.Cancelled as e:
    raise ValueError() from e  # error: 4


def foo():
    try:
        ...
    except trio.Cancelled:
        raise
    except BaseException:
        return  # would otherwise error
    except:
        return  # would otherwise error

    try:
        ...
    except trio.Cancelled:
        raise
    except:
        return  # would otherwise error



"""Submodule for all error classes/visitors.

Exports ERROR_CLASSES and default_disabled_error_codes to be used by others, and populates
them by importing all files with visitor classes (which needs to manually maintained).
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .flake8triovisitor import Flake8TrioVisitor

__all__ = ["ERROR_CLASSES", "default_disabled_error_codes", "utility_visitors"]
ERROR_CLASSES: set[type[Flake8TrioVisitor]] = set()
utility_visitors: set[type[Flake8TrioVisitor]] = set()
default_disabled_error_codes: list[str] = []

# Import all visitors so their decorators run, filling the above containers
# This has to be done at the end to avoid circular imports
from . import visitor_utility  # isort: skip
from . import visitor102  # isort: skip
from . import visitor103_104  # isort: skip
from . import visitor105  # isort: skip
from . import visitor111  # isort: skip
from . import visitor118  # isort: skip
from . import visitor2xx  # isort: skip
from . import visitor91x  # isort: skip
from . import visitors  # isort: skip



# type: ignore
import anyio

try:
    ...
except anyio.get_cancelled_exc_class() as e:
    raise ValueError() from e  # error: 4

# anyio.Cancelled does not exist
try:
    ...
except anyio.Cancelled as e:
    raise ValueError() from e


def foo():
    try:
        ...
    except anyio.get_cancelled_exc_class():
        raise
    except BaseException:
        return  # would otherwise error
    except:
        return  # would otherwise error

    try:
        ...
    except anyio.get_cancelled_exc_class():
        raise
    except:
        return  # would otherwise error



# type: ignore
import os.path
from os.path import isfile, normpath, relpath


async def foo():
    normpath("")  # TRIO240: 4, 'normpath', "trio"
    relpath("")  # TRIO240: 4, 'relpath', "trio"
    isfile("")  # TRIO240: 4, 'isfile', "trio"

    os.path._path_normpath(...)  # TRIO240: 4, '_path_normpath', "trio"
    os.path.normpath("")  # TRIO240: 4, 'normpath', "trio"
    os.path._joinrealpath(...)  # TRIO240: 4, '_joinrealpath', "trio"
    os.path.islink("")  # TRIO240: 4, 'islink', "trio"
    os.path.lexists("")  # TRIO240: 4, 'lexists', "trio"
    os.path.ismount("")  # TRIO240: 4, 'ismount', "trio"
    os.path.realpath("")  # TRIO240: 4, 'realpath', "trio"
    os.path.exists("")  # TRIO240: 4, 'exists', "trio"
    os.path.isdir("")  # TRIO240: 4, 'isdir', "trio"
    os.path.isfile("")  # TRIO240: 4, 'isfile', "trio"
    os.path.getatime("")  # TRIO240: 4, 'getatime', "trio"
    os.path.getctime("")  # TRIO240: 4, 'getctime', "trio"
    os.path.getmtime("")  # TRIO240: 4, 'getmtime', "trio"
    os.path.getsize("")  # TRIO240: 4, 'getsize', "trio"
    os.path.samefile("", "")  # TRIO240: 4, 'samefile', "trio"
    os.path.sameopenfile(0, 1)  # TRIO240: 4, 'sameopenfile', "trio"
    os.path.relpath("")  # TRIO240: 4, 'relpath', "trio"

    await os.path.isfile("")  # TRIO240: 10, 'isfile', "trio"
    print(os.path.isfile(""))  # TRIO240: 10, 'isfile', "trio"
    os.path.abspath("")
    os.path.anything("")


def foo2():
    os.path.isfile("")



# type: ignore
import io
from io import BufferedRandom, BufferedReader, BufferedWriter, TextIOWrapper
from typing import Any, Optional

blah: Any = None


async def file_text(f: io.TextIOWrapper):
    f.read()  # TRIO232: 4, 'read', 'f', "trio"
    f.readlines()  # TRIO232: 4, 'readlines', 'f', "trio"

    # there might be non-sync calls on TextIOWrappers? - but it will currently trigger
    # on all calls
    f.anything()  # TRIO232: 4, 'anything', 'f', "trio"


async def file_binary_read(f: io.BufferedReader):
    f.read()  # TRIO232: 4, 'read', 'f', "trio"


async def file_binary_write(f: io.BufferedWriter):
    f.read()  # TRIO232: 4, 'read', 'f', "trio"


async def file_binary_readwrite(f: io.BufferedRandom):
    f.read()  # TRIO232: 4, 'read', 'f', "trio"


async def file_text_(f: TextIOWrapper):
    f.read()  # TRIO232: 4, 'read', 'f', "trio"


async def file_binary_read_(f: BufferedReader):
    f.read()  # TRIO232: 4, 'read', 'f', "trio"


async def file_binary_write_(f: BufferedWriter):
    f.read()  # TRIO232: 4, 'read', 'f', "trio"


async def file_binary_readwrite_(f: BufferedRandom):
    f.read()  # TRIO232: 4, 'read', 'f', "trio"


async def file_text_3(f: TextIOWrapper = blah):
    f.read()  # TRIO232: 4, 'read', 'f', "trio"


async def file_text_4(f: TextIOWrapper | None):
    f.read()  # TRIO232: 4, 'read', 'f', "trio"
    if f:
        f.read()  # TRIO232: 8, 'read', 'f', "trio"


async def file_text_4_left(f: None | TextIOWrapper):
    f.read()  # TRIO232: 4, 'read', 'f', "trio"
    if f:
        f.read()  # TRIO232: 8, 'read', 'f', "trio"


# not handled
async def file_text_4_both(f: None | TextIOWrapper | None):
    f.read()


# not handled
async def file_text_4_non_none(f: TextIOWrapper | int):
    f.read()


async def file_text_5(f: TextIOWrapper | None = None):
    f.read()  # TRIO232: 4, 'read', 'f', "trio"
    if f:
        f.read()  # TRIO232: 8, 'read', 'f', "trio"


async def file_text_6(f: Optional[TextIOWrapper] = None):
    f.read()  # TRIO232: 4, 'read', 'f', "trio"
    if f:
        f.read()  # TRIO232: 8, 'read', 'f', "trio"


# posonly
async def file_text_7(f: TextIOWrapper = blah, /):
    f.read()  # TRIO232: 4, 'read', 'f', "trio"


# keyword-only
async def file_text_8(*, f: TextIOWrapper = blah):
    f.read()  # TRIO232: 4, 'read', 'f', "trio"


async def file_text_9(lf: list[TextIOWrapper]):
    for f in lf:
        f.read()  # not handled


async def open_file_1():
    with open(""):
        ...


async def open_file_2():
    with open("") as f:
        print(f)


async def open_file_3():
    with open("") as f:
        f.read()  # TRIO232: 8, 'read', 'f', "trio"


async def open_file_4():
    f = open("")


async def open_file_5():
    f = open("")
    f.read()  # TRIO232: 4, 'read', 'f', "trio"


async def open_file_6():
    ff = open("")
    f = ff
    f.read()  # TRIO232: 4, 'read', 'f', "trio"
    ff.read()  # TRIO232: 4, 'read', 'ff', "trio"


async def noerror():
    with blah("") as f:
        f.read()


def sync_fun(f: TextIOWrapper):
    async def async_fun():
        f.read()  # TRIO232: 8, 'read', 'f', "trio"


def sync_fun_2():
    f = open("")

    async def async_fun():
        f.read()  # TRIO232: 8, 'read', 'f', "trio"


async def type_assign():
    f: TextIOWrapper = ...
    f.read()  # TRIO232: 4, 'read', 'f', "trio"


# define global variables
f = open("")
g: TextIOWrapper = ...


# Test state handling on nested functions
async def async_wrapper(f: TextIOWrapper):
    def inside(f: int):
        f.read()

        async def inception():
            f.read()

        f.read()

    f.read()  # TRIO232: 4, 'read', 'f', "trio"
    lambda f: f.read()
    f.read()  # TRIO232: 4, 'read', 'f', "trio"
    f.read()  # TRIO232: 4, 'read', 'f', "trio"


# and show that they're still marked as TextIOWrappers
async def global_vars():
    f.read()  # TRIO232: 4, 'read', 'f', "trio"
    g.read()  # TRIO232: 4, 'read', 'g', "trio"


# If the type is explicitly overridden, it will not error
async def overridden_type(f: TextIOWrapper):
    f: int = 7
    f.read()
    f: TextIOWrapper = ...
    f.read()  # TRIO232: 4, 'read', 'f', "trio"
    f: int = 7
    f.read()
    f: TextIOWrapper = ...
    f.read()  # TRIO232: 4, 'read', 'f', "trio"


# ***** Known unhandled cases *****


# It will error on non-explicit assignments
async def implicit_overridden_type():
    f: TextIOWrapper = ...
    f = arbitrary_function()
    f.read()  # TRIO232: 4, 'read', 'f', "trio"
    f = 7
    f.read()  # TRIO232: 4, 'read', 'f', "trio"


# Tuple assignments are completely ignored
async def multi_assign():
    x, y = open(""), open("")
    x.read()  # should error
    y.read()  # should error


# as are attribute assignments
async def attribute_assign():
    x.y = open("")
    x.y.read()  # should error


# or calling on subattributes of an object
async def attribute_access_on_object():
    f: TextIOWrapper = ...
    f.any.thing()  # should error


# The type checker is very naive, and will not do any parsing of logic pertaining
# to the type
async def type_restricting_1(f: Optional[TextIOWrapper] = None):
    if f is None:
        f.read()  # TRIO232: 8, 'read', 'f', "trio"


async def type_restricting_2(f: Optional[TextIOWrapper] = None):
    if isinstance(f, TextIOWrapper):
        return
    f.read()  # TRIO232: 4, 'read', 'f', "trio"


# Classes are not supported, partly due to not handling attributes at all,
# but would require additional logic
class myclass:
    x: TextIOWrapper

    async def foo(self):
        self.x.read()  # should error


class myclass_2:
    def __init__(self):
        self.x: TextIOWrapper = open("")

    async def foo(self):
        self.x.read()  # should error


# Return types are not parsed
async def call_function_with_return_type():
    def return_textiowrapper() -> TextIOWrapper:
        return open("")

    k = return_textiowrapper()
    k.read()  # should error



"""2XX error classes, which checks for blocking sync calls in async functions.

200 is user-configured, doing nothing by default.
210 looks for usage of HTTP requests from common http libraries.
211 additionally matches on object methods whose signature looks like an http request.
220&221 looks for subprocess and os calls that should be wrapped.
230&231 looks for os.open and os.fdopen that should be wrapped.
240 looks for os.path functions that interact with the disk in various ways.
"""

from __future__ import annotations

import ast
import re
from typing import Any

from .flake8triovisitor import Flake8TrioVisitor
from .helpers import error_class, fnmatch_qualified_name, get_matching_call


@error_class
class Visitor200(Flake8TrioVisitor):
    error_codes = {
        "TRIO200": "User-configured blocking sync call {0} in async function, consider "
        "replacing with {1}.",
    }

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.async_function = False

    def visit_AsyncFunctionDef(
        self, node: ast.AsyncFunctionDef | ast.FunctionDef | ast.Lambda
    ):
        self.save_state(node, "async_function")
        self.async_function = isinstance(node, ast.AsyncFunctionDef)

    visit_FunctionDef = visit_AsyncFunctionDef
    visit_Lambda = visit_AsyncFunctionDef

    def visit_Call(self, node: ast.Call):
        if self.async_function and not getattr(node, "awaited", False):
            self.visit_blocking_call(node)

    def visit_blocking_call(self, node: ast.Call):
        blocking_calls = self.options.trio200_blocking_calls
        if key := fnmatch_qualified_name([node.func], *blocking_calls):
            self.error(node, key, blocking_calls[key])


# used by Visitor212 and Visitor232


@error_class
class Visitor21X(Visitor200):
    error_codes = {
        "TRIO210": "Sync HTTP call {} in async function, use `httpx.AsyncClient`.",
        "TRIO211": (
            "Likely sync HTTP call {} in async function, use `httpx.AsyncClient`."
        ),
    }

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.imports: set[str] = set()

    def visit_ImportFrom(self, node: ast.ImportFrom):
        if node.module == "urllib3":
            self.imports.add(node.module)

    def visit_Import(self, node: ast.Import):
        for name in node.names:
            if name.name == "urllib3":
                # Could also save the name.asname for matching
                self.imports.add(name.name)

    def visit_blocking_call(self, node: ast.Call):
        http_methods = {
            "get",
            "options",
            "head",
            "post",
            "put",
            "patch",
            "delete",
        }
        func_name = ast.unparse(node.func)
        for http_package in "requests", "httpx":
            if get_matching_call(node, *http_methods | {"request"}, base=http_package):
                self.error(node, func_name, error_code="TRIO210")
                return

        if func_name in (
            "urllib3.request",
            "urllib.request.urlopen",
            "request.urlopen",
            "urlopen",
        ):
            self.error(node, func_name, error_code="TRIO210")

        elif (
            "urllib3" in self.imports
            and isinstance(node.func, ast.Attribute)
            and node.func.attr == "request"
            and node.args
            and isinstance(node.args[0], ast.Constant)
            and isinstance(node.args[0].value, str)
            and node.args[0].value.lower() in http_methods | {"trace", "connect"}
        ):
            self.error(node, func_name, error_code="TRIO211")


@error_class
class Visitor212(Visitor200):
    error_codes = {
        "TRIO212": (
            "Blocking sync HTTP call {1} on httpx object {0}, use httpx.AsyncClient."
        )
    }

    def __init__(self, *args: Any, **kwargs: Any):
        self.dangerous_classes = (
            "httpx.Client",
            "urllib3.HTTPConnectionPool",
            "urllib3.HTTPSConnectionPool",
            "urllib3.PoolManager",
            "urllib3.ProxyManager",
            "urllib3.connectionpool.ConnectionPool",
            "urllib3.connectionpool.HTTPConnectionPool",
            "urllib3.connectionpool.HTTPSConnectionPool",
            "urllib3.poolmanager.PoolManager",
            "urllib3.poolmanager.ProxyManager",
            "urllib3.request.RequestMethods",
        )
        super().__init__(*args, **kwargs)
        for c in self.dangerous_classes:
            self.typed_calls[c] = c

    def visit_blocking_call(self, node: ast.Call):
        httpx_blocking_methods = (
            "close",
            "delete",
            "get",
            "head",
            "options",
            "patch",
            "post",
            "put",
            "request",
            "send",
            "stream",
        )
        if (
            isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and node.func.value.id in self.variables
            and (anno := self.variables.get(node.func.value.id))
            and (anno != "httpx.Client" or node.func.attr in httpx_blocking_methods)
            and anno in self.dangerous_classes
        ):
            self.error(node, node.func.attr, node.func.value.id)


# Process invocations 202
@error_class
class Visitor22X(Visitor200):
    error_codes = {
        "TRIO220": (
            "Sync call {} in async function, use "
            "`await nursery.start({}.run_process, ...)`."
        ),
        "TRIO221": "Sync call {} in async function, use `await {}.run_process(...)`.",
        "TRIO222": "Sync call {} in async function, wrap in `{}.to_thread.run_sync()`.",
    }

    def visit_blocking_call(self, node: ast.Call):
        def is_p_wait(arg: ast.expr) -> bool:
            return (isinstance(arg, ast.Attribute) and arg.attr == "P_WAIT") or (
                isinstance(arg, ast.Name) and arg.id == "P_WAIT"
            )

        subprocess_calls = {
            "run",
            "call",
            "check_call",
            "check_output",
            "getoutput",
            "getstatusoutput",
        }

        func_name = ast.unparse(node.func)
        error_code: str | None = None
        if func_name in ("subprocess.Popen", "os.popen"):
            error_code = "TRIO220"

        elif func_name in (
            "os.system",
            "os.posix_spawn",
            "os.posix_spawnp",
        ) or get_matching_call(node, *subprocess_calls, base="subprocess"):
            error_code = "TRIO221"

        elif re.fullmatch("os.wait([34]|(id)|(pid))?", func_name):
            error_code = "TRIO222"

        elif re.fullmatch("os.spawn[vl]p?e?", func_name):
            error_code = "TRIO221"

            # if mode= is given and not [os.]P_WAIT: TRIO220
            # 1. as a positional parameter
            if node.args and not is_p_wait(node.args[0]):
                error_code = "TRIO220"

            # 2. as a keyword parameter
            for kw in node.keywords:
                if kw.arg == "mode" and not is_p_wait(kw.value):
                    error_code = "TRIO220"
                    break

        if error_code is not None:
            self.error(node, func_name, self.library_str, error_code=error_code)


@error_class
class Visitor23X(Visitor200):
    error_codes = {
        "TRIO230": ("Sync call {0} in async function, use `{1}.open_file(...)`."),
        "TRIO231": ("Sync call {0} in async function, use `{1}.wrap_file({0})`."),
    }

    def visit_Call(self, node: ast.Call):
        func_name = ast.unparse(node.func)
        if re.fullmatch(r"(trio|anyio)\.wrap_file", func_name) and len(node.args) == 1:
            setattr(node.args[0], "wrapped", True)  # noqa: B010
        super().visit_Call(node)

    def visit_blocking_call(self, node: ast.Call):
        if getattr(node, "wrapped", False):
            return
        func_name = ast.unparse(node.func)
        if func_name in ("open", "io.open", "io.open_code"):
            error_code = "TRIO230"
        elif func_name == "os.fdopen":
            error_code = "TRIO231"
        else:
            return
        self.error(node, func_name, self.library_str, error_code=error_code)


@error_class
class Visitor232(Visitor200):
    error_codes = {
        "TRIO232": (
            "Blocking sync call {1} on file object {0}, wrap the file object"
            "in `{2}.wrap_file()` to get an async file object."
        )
    }

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.typed_calls["open"] = "io.TextIOWrapper"

    def visit_blocking_call(self, node: ast.Call):
        io_file_types = (
            "io.TextIOWrapper",
            "io.BufferedReader",
            "io.BufferedWriter",
            "io.BufferedRandom",
        )
        if (
            isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and node.func.value.id in self.variables
            and (anno := self.variables.get(node.func.value.id))
            and (anno in io_file_types or f"io.{anno}" in io_file_types)
        ):
            self.error(node, node.func.attr, node.func.value.id, self.library_str)


@error_class
class Visitor24X(Visitor200):
    error_codes = {
        "TRIO240": ("Avoid using os.path, prefer using {1}.Path objects."),
    }

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.imports_from_ospath: set[str] = set()

    os_funcs = (
        "_path_normpath",
        "normpath",
        "_joinrealpath",
        "islink",
        "lexists",
        "ismount",  # safe on windows, unsafe on posix
        "realpath",
        "exists",
        "isdir",
        "isfile",
        "getatime",
        "getctime",
        "getmtime",
        "getsize",
        "samefile",
        "sameopenfile",
        "relpath",
    )

    # doesn't protect against `from os import path` or `import os.path as <x>`
    # but those should be very rare
    def visit_ImportFrom(self, node: ast.ImportFrom):
        if node.module == "os.path":
            for alias in node.names:
                self.imports_from_ospath.add(
                    alias.asname if alias.asname is not None else alias.name
                )

    def visit_Call(self, node: ast.Call):
        if not self.async_function:
            return
        func_name = ast.unparse(node.func)
        if func_name in self.imports_from_ospath:
            self.error(node, func_name, self.library_str)
        elif (m := re.fullmatch(r"os\.path\.(?P<func>.*)", func_name)) and m.group(
            "func"
        ) in self.os_funcs:
            self.error(node, m.group("func"), self.library_str)



"""Visitor102, which warns on unprotected `await` inside `finally`.

To properly protect they must be inside a shielded cancel scope with a timeout.
"""

from __future__ import annotations

import ast
from typing import Any

from ..base import Statement
from .flake8triovisitor import Flake8TrioVisitor
from .helpers import cancel_scope_names, critical_except, error_class, get_matching_call


@error_class
class Visitor102(Flake8TrioVisitor):
    error_codes = {
        "TRIO102": (
            "await inside {0.name} on line {0.lineno} must have shielded cancel "
            "scope with a timeout."
        ),
    }

    class TrioScope:
        def __init__(self, node: ast.Call, funcname: str, _):
            self.node = node
            self.funcname = funcname
            self.variable_name: str | None = None
            self.shielded: bool = False
            self.has_timeout: bool = True

            # scope.shielded is assigned to in visit_Assign

            if self.funcname == "CancelScope":
                self.has_timeout = False
                for kw in node.keywords:
                    # Only accepts constant values
                    if kw.arg == "shield" and isinstance(kw.value, ast.Constant):
                        self.shielded = kw.value.value
                    # sets to True even if timeout is explicitly set to inf
                    if kw.arg == "deadline":
                        self.has_timeout = True

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._critical_scope: Statement | None = None
        self._trio_context_managers: list[Visitor102.TrioScope] = []
        self.cancelled_caught = False

    # if we're inside a finally, and we're not inside a scope that doesn't have
    # both a timeout and shield
    def visit_Await(self, node: ast.Await | ast.AsyncFor | ast.AsyncWith):
        if self._critical_scope is not None and not any(
            cm.has_timeout and cm.shielded for cm in self._trio_context_managers
        ):
            self.error(node, self._critical_scope)

    visit_AsyncFor = visit_Await

    def visit_With(self, node: ast.With | ast.AsyncWith):
        self.save_state(node, "_trio_context_managers", copy=True)

        # Check for a `with trio.<scope_creator>`
        for item in node.items:
            call = get_matching_call(
                item.context_expr, "open_nursery", *cancel_scope_names
            )
            if call is None:
                continue

            trio_scope = self.TrioScope(*call)
            # check if it's saved in a variable
            if isinstance(item.optional_vars, ast.Name):
                trio_scope.variable_name = item.optional_vars.id

            self._trio_context_managers.append(trio_scope)
            break

    def visit_AsyncWith(self, node: ast.AsyncWith):
        self.visit_Await(node)
        self.visit_With(node)

    def visit_Try(self, node: ast.Try):
        self.save_state(
            node, "_critical_scope", "_trio_context_managers", "cancelled_caught"
        )
        self.cancelled_caught = False
        # There's no visit_Finally, so we need to manually visit the Try fields.
        self.visit_nodes(node.body, node.handlers, node.orelse)

        self._trio_context_managers = []
        self._critical_scope = Statement("try/finally", node.lineno, node.col_offset)
        self.visit_nodes(node.finalbody)

    def visit_ExceptHandler(self, node: ast.ExceptHandler):
        if self.cancelled_caught:
            return
        res = critical_except(node)
        if res is None:
            return

        self.save_state(node, "_critical_scope", "_trio_context_managers")
        self.cancelled_caught = True
        self._trio_context_managers = []
        self._critical_scope = res

    def visit_Assign(self, node: ast.Assign):
        # checks for <scopename>.shield = [True/False]
        if self._trio_context_managers and len(node.targets) == 1:
            last_scope = self._trio_context_managers[-1]
            target = node.targets[0]
            if (
                last_scope.variable_name is not None
                and isinstance(target, ast.Attribute)
                and isinstance(target.value, ast.Name)
                and target.value.id == last_scope.variable_name
                and target.attr == "shield"
                and isinstance(node.value, ast.Constant)
            ):
                last_scope.shielded = node.value.value



# ARG --enable-visitor-codes-regex=(TRIO103)|(TRIO104)
# NOANYIO

from typing import Any

import trio


def foo() -> Any:
    ...


# fmt: off
# TODO: Black 23.1.0 moves the long comments around a bit.

try:
    ...
except (
    SyntaxError,
    ValueError,
    trio.Cancelled,  # TRIO103: 4, "trio.Cancelled"
) as p:
    ...

try:
    ...
except (SyntaxError, ValueError):
    raise

try:
    ...
except trio.Cancelled as e:
    raise e

try:
    ...
except trio.Cancelled as e:
    raise  # acceptable - see https://peps.python.org/pep-0678/#example-usage

try:
    ...
except trio.Cancelled:  # TRIO103: 7, "trio.Cancelled"
    ...

# Issue #106, false alarm on excepts after `Cancelled` has already been handled
try:
    ...
except trio.Cancelled:
    raise
except BaseException:  # now silent
    ...
except:  # now silent
    ...

try:
    ...
except trio.Cancelled:
    raise
except:  # now silent
    ...


try:
    ...
except BaseException:
    raise
except:  # now silent
    ...

# don't throw multiple 103's even if `Cancelled` wasn't properly handled.
try:
    ...
except trio.Cancelled:  # TRIO103: 7, "trio.Cancelled"
    ...
except BaseException:  # now silent
    ...
except:  # now silent
    ...

# Check state management of whether cancelled has been handled across nested try's
try:
    try:
        ...
    except trio.Cancelled:
        raise
except trio.Cancelled:  # TRIO103: 7, "trio.Cancelled"
    ...
except:
    try:
        ...
    except trio.Cancelled:  # TRIO103: 11, "trio.Cancelled"
        ...



"""Contains the base class that all error classes inherit from."""

from __future__ import annotations

import ast
import re
from abc import ABC
from typing import TYPE_CHECKING, Any, Union

from ..base import Error, Statement

if TYPE_CHECKING:
    from collections.abc import Iterable

    from ..runner import SharedState

    HasLineCol = Union[
        ast.expr, ast.stmt, ast.arg, ast.excepthandler, ast.alias, Statement
    ]


class Flake8TrioVisitor(ast.NodeVisitor, ABC):
    # abstract attribute by not providing a value
    error_codes: dict[str, str]

    def __init__(self, shared_state: SharedState):
        super().__init__()
        assert self.error_codes, "subclass must define error_codes"
        self.outer: dict[ast.AST, dict[str, Any]] = {}
        self.novisit = False
        self.__state = shared_state

        self.options = self.__state.options
        self.typed_calls = self.__state.typed_calls

        # mark variables that shouldn't be saved/loaded in self.get_state
        self.nocopy = {
            "_Flake8TrioVisitor__state",
            "error_codes",
            "nocopy",
            "novisit",
            "options",
            "outer",
            "typed_calls",
        }

        self.suppress_errors = False

    # `variables` can be saved/loaded, but need a setter to not clear the reference
    @property
    def variables(self):
        return self.__state.variables

    @variables.setter
    def variables(self, value):
        self.__state.variables.clear()
        self.__state.variables.update(value)

    def visit(self, node: ast.AST):
        """Visit a node."""
        # construct visitor for this node type
        visitor = getattr(self, "visit_" + node.__class__.__name__, None)

        # if we have a visitor for it, visit it
        # it will set self.novisit if it manually visits children
        self.novisit = False

        if visitor is not None:
            visitor(node)

        # if it doesn't set novisit, let the regular NodeVisitor iterate through
        # subfields
        if not self.novisit:
            super().generic_visit(node)

        # if an outer state was saved in this node restore it after visiting children
        self.set_state(self.outer.pop(node, {}))

        # set novisit so external runner doesn't visit this node with this class
        self.novisit = True

    def visit_nodes(self, *nodes: ast.AST | Iterable[ast.AST]):
        for arg in nodes:
            if isinstance(arg, ast.AST):
                self.visit(arg)
            else:
                for node in arg:
                    self.visit(node)

    def error(
        self,
        node: HasLineCol,
        *args: str | Statement | int,
        error_code: str | None = None,
    ):
        if error_code is None:
            assert (
                len(self.error_codes) == 1
            ), "No error code defined, but class has multiple codes"
            error_code = next(iter(self.error_codes))
        # don't emit an error if this code is disabled in a multi-code visitor
        elif not re.match(self.options.enable_visitor_codes_regex, error_code):
            return

        if not self.suppress_errors:
            self.__state.problems.append(
                Error(
                    # 7 == len('TRIO...'), so alt messages raise the original code
                    error_code[:7],
                    node.lineno,
                    node.col_offset,
                    self.error_codes[error_code],
                    *args,
                )
            )

    def get_state(self, *attrs: str, copy: bool = False) -> dict[str, Any]:
        if not attrs:
            # get all attributes, unless they're marked as nocopy
            attrs = tuple(set(self.__dict__.keys()) - self.nocopy)
        res: dict[str, Any] = {}
        for attr in attrs:
            value = getattr(self, attr)
            if copy and hasattr(value, "copy"):
                value = value.copy()
            res[attr] = value
        return res

    def set_state(self, attrs: dict[str, Any], copy: bool = False):
        for attr, value in attrs.items():
            if copy and hasattr(value, "copy"):
                value = value.copy()
            setattr(self, attr, value)

    def save_state(self, node: ast.AST, *attrs: str, copy: bool = False):
        state = self.get_state(*attrs, copy=copy)
        if node in self.outer:
            # not currently used, and not gonna bother adding dedicated test
            # visitors atm
            self.outer[node].update(state)  # pragma: no cover
        else:
            self.outer[node] = state

    def walk(self, *body: ast.AST) -> Iterable[ast.AST]:
        for b in body:
            yield from ast.walk(b)

    @property
    def library(self) -> tuple[str, ...]:
        return self.__state.library if self.__state.library else ("trio",)

    @property
    def library_str(self) -> str:
        if len(self.library) == 1:
            return self.library[0]
        return "[" + "|".join(self.library) + "]"

    def add_library(self, name) -> None:
        if name not in self.__state.library:
            self.__state.library = self.__state.library + (name,)



"""Contains base classes used across multiple parts of the package."""

from __future__ import annotations

from typing import Any, NamedTuple


class Statement(NamedTuple):
    name: str
    lineno: int
    col_offset: int = -1

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, Statement)
            and self[:2] == other[:2]  # type: ignore
            and (
                self.col_offset == other.col_offset
                or -1 in (self.col_offset, other.col_offset)
            )
        )


class Error:
    def __init__(
        self, error_code: str, lineno: int, col: int, message: str, *args: object
    ):
        self.line = lineno
        self.col = col
        self.code = error_code
        self.message = message
        self.args = args

    # for yielding to flake8
    def __iter__(self):
        yield self.line
        yield self.col
        yield f"{self.code} " + self.message.format(*self.args)
        # We are no longer yielding `type(Plugin)` since that's quite tricky to do
        # without circular imports, and afaik flake8 doesn't care anymore.
        yield None

    def cmp(self):
        return self.line, self.col, self.code, self.args

    # for sorting in tests
    def __lt__(self, other: Any) -> bool:
        assert isinstance(other, Error)
        return self.cmp() < other.cmp()

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Error) and self.cmp() == other.cmp()

    def __repr__(self) -> str:  # pragma: no cover
        trailer = "".join(f", {x!r}" for x in self.args)
        return f"<{self.code} error at {self.line}:{self.col}{trailer}>"



"""Contains Visitor111 which looks for incorrectly nested nurseries."""

from __future__ import annotations

import ast
from typing import Any, NamedTuple

from .flake8triovisitor import Flake8TrioVisitor
from .helpers import error_class, get_matching_call


@error_class
class Visitor111(Flake8TrioVisitor):
    error_codes = {
        "TRIO111": (
            "variable {2} is usable within the context manager on line {0}, but that "
            "will close before nursery opened on line {1} - this is usually a bug.  "
            "Nurseries should generally be the inner-most context manager."
        ),
    }

    class NurseryCall(NamedTuple):
        stack_index: int
        name: str

    class TrioContextManager(NamedTuple):
        lineno: int
        name: str
        is_nursery: bool

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._context_managers: list[Visitor111.TrioContextManager] = []
        self._nursery_call: Visitor111.NurseryCall | None = None

    def visit_With(self, node: ast.With | ast.AsyncWith):
        self.save_state(node, "_context_managers", copy=True)
        for item in node.items:
            # 111
            # if a withitem is saved in a variable,
            # push its line, variable, and whether it's a trio nursery
            # to the _context_managers stack,
            if isinstance(item.optional_vars, ast.Name):
                self._context_managers.append(
                    self.TrioContextManager(
                        item.context_expr.lineno,
                        item.optional_vars.id,
                        get_matching_call(item.context_expr, "open_nursery")
                        is not None,
                    )
                )

    visit_AsyncWith = visit_With

    def visit_FunctionDef(
        self, node: ast.FunctionDef | ast.AsyncFunctionDef | ast.Lambda
    ):
        self.save_state(node)
        self._context_managers = []
        self._nursery_call = None

    visit_AsyncFunctionDef = visit_FunctionDef
    visit_Lambda = visit_FunctionDef

    # if it's a <X>.start[_soon] call
    # and <X> is a nursery listed in self._context_managers:
    # Save <X>'s index in self._context_managers to guard against cm's higher in the
    # stack being passed as parameters to it. (and save <X> for the error message)
    def visit_Call(self, node: ast.Call):
        self.save_state(node, "_nursery_call")

        if (
            isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and node.func.attr in ("start", "start_soon")
        ):
            self._nursery_call = None
            for i, cm in enumerate(self._context_managers):
                if node.func.value.id == cm.name:
                    # don't break upon finding a nursery in case there's multiple cm's
                    # on the stack with the same name
                    if cm.is_nursery:
                        self._nursery_call = self.NurseryCall(i, node.func.attr)
                    else:
                        self._nursery_call = None

    # If we're inside a <X>.start[_soon] call (where <X> is a nursery),
    # and we're accessing a variable cm that's on the self._context_managers stack,
    # with a higher index than <X>:
    #   Raise error since the scope of cm may close before the function passed to the
    # nursery finishes.
    def visit_Name(self, node: ast.Name):
        if self._nursery_call is None:
            return

        for i, cm in enumerate(self._context_managers):
            if cm.name == node.id and i > self._nursery_call.stack_index:
                self.error(
                    node,
                    cm.lineno,
                    self._context_managers[self._nursery_call.stack_index].lineno,
                    node.id,
                    self._nursery_call.name,
                )



"""Contains Visitor91X.

910 looks for async functions without guaranteed checkpoints (or exceptions), and 911 does
the same except for async iterables - while also requiring that they checkpoint between
each yield.
"""

from __future__ import annotations

import ast
from typing import Any

from ..base import Statement
from .flake8triovisitor import Flake8TrioVisitor
from .helpers import (
    disabled_by_default,
    error_class,
    fnmatch_qualified_name,
    has_decorator,
    iter_guaranteed_once,
)


# used in 910/911
def empty_body(body: list[ast.stmt]) -> bool:
    # Does the function body consist solely of `pass`, `...`, and (doc)string literals?
    return all(
        isinstance(stmt, ast.Pass)
        or (
            isinstance(stmt, ast.Expr)
            and isinstance(stmt.value, ast.Constant)
            and (stmt.value.value is Ellipsis or isinstance(stmt.value.value, str))
        )
        for stmt in body
    )


@error_class
@disabled_by_default
class Visitor91X(Flake8TrioVisitor):
    error_codes = {
        "TRIO910": (
            "{0} from async function with no guaranteed checkpoint or exception "
            "since function definition on line {1.lineno}."
        ),
        "TRIO911": (
            "{0} from async iterable with no guaranteed checkpoint since {1.name} "
            "on line {1.lineno}."
        ),
    }

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.has_yield = False
        self.safe_decorator = False
        self.async_function = False

        self.uncheckpointed_statements: set[Statement] = set()
        self.uncheckpointed_before_continue: set[Statement] = set()
        self.uncheckpointed_before_break: set[Statement] = set()

        self.default = self.get_state()

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef):
        # don't lint functions whose bodies solely consist of pass or ellipsis
        if has_decorator(node.decorator_list, "overload") or empty_body(node.body):
            return

        self.save_state(node)
        self.set_state(self.default, copy=True)

        # disable checks in asynccontextmanagers by saying the function isn't async
        self.async_function = not fnmatch_qualified_name(
            node.decorator_list, *self.options.no_checkpoint_warning_decorators
        )
        if not self.async_function:
            return

        self.uncheckpointed_statements = {
            Statement("function definition", node.lineno, node.col_offset)
        }

        self.generic_visit(node)

        self.check_function_exit(node)

    # error if function exits or returns with uncheckpointed statements
    def check_function_exit(self, node: ast.Return | ast.AsyncFunctionDef):
        for statement in self.uncheckpointed_statements:
            self.error(
                node,
                "return" if isinstance(node, ast.Return) else "exit",
                statement,
                error_code="TRIO911" if self.has_yield else "TRIO910",
            )

    def visit_Return(self, node: ast.Return):
        if not self.async_function:
            return
        self.generic_visit(node)
        self.check_function_exit(node)

        # avoid duplicate error messages
        self.uncheckpointed_statements = set()

    # disregard checkpoints in nested function definitions
    def visit_FunctionDef(self, node: ast.FunctionDef):
        self.save_state(node)
        self.set_state(self.default, copy=True)

    # checkpoint functions
    def visit_Await(self, node: ast.Await | ast.Raise):
        # the expression being awaited is not checkpointed
        # so only set checkpoint after the await node
        self.generic_visit(node)

        # all nodes are now checkpointed
        self.uncheckpointed_statements = set()

    # raising exception means we don't need to checkpoint so we can treat it as one
    visit_Raise = visit_Await

    # Async context managers can reasonably checkpoint on either or both of entry and
    # exit.  Given that we can't tell which, we assume "both" to avoid raising a
    # missing-checkpoint warning when there might in fact be one (i.e. a false alarm).
    def visit_AsyncWith(self, node: ast.AsyncWith):
        self.visit_nodes(node.items)

        self.uncheckpointed_statements = set()
        self.visit_nodes(node.body)

        self.uncheckpointed_statements = set()

    # error if no checkpoint since earlier yield or function entry
    def visit_Yield(self, node: ast.Yield):
        if not self.async_function:
            return
        self.has_yield = True
        self.generic_visit(node)
        for statement in self.uncheckpointed_statements:
            self.error(
                node,
                "yield",
                statement,
                error_code="TRIO911",
            )

        # mark as requiring checkpoint after
        self.uncheckpointed_statements = {
            Statement("yield", node.lineno, node.col_offset)
        }

    # valid checkpoint if there's valid checkpoints (or raise) in:
    # (try or else) and all excepts, or in finally
    #
    # try can jump into any except or into the finally* at any point during it's
    # execution so we need to make sure except & finally can handle worst-case
    # * unless there's a bare except / except BaseException - not implemented.
    def visit_Try(self, node: ast.Try):
        if not self.async_function:
            return
        # except & finally guaranteed to enter with checkpoint if checkpointed
        # before try and no yield in try body.
        body_uncheckpointed_statements = self.uncheckpointed_statements.copy()
        for inner_node in self.walk(*node.body):
            if isinstance(inner_node, ast.Yield):
                body_uncheckpointed_statements.add(
                    Statement("yield", inner_node.lineno, inner_node.col_offset)
                )

        # check try body
        self.visit_nodes(node.body)

        # save state at end of try for entering else
        try_checkpoint = self.uncheckpointed_statements

        # check that all except handlers checkpoint (await or most likely raise)
        except_uncheckpointed_statements: set[Statement] = set()

        for handler in node.handlers:
            # enter with worst case of try
            self.uncheckpointed_statements = body_uncheckpointed_statements.copy()

            self.visit_nodes(handler)

            except_uncheckpointed_statements.update(self.uncheckpointed_statements)

        # check else
        # if else runs it's after all of try, so restore state to back then
        self.uncheckpointed_statements = try_checkpoint
        self.visit_nodes(node.orelse)

        # checkpoint if else checkpoints, and all excepts checkpoint
        self.uncheckpointed_statements.update(except_uncheckpointed_statements)

        if node.finalbody:
            added = body_uncheckpointed_statements.difference(
                self.uncheckpointed_statements
            )
            # if there's no bare except or except BaseException, we can jump into
            # finally from any point in try. But the exception will be reraised after
            # finally, so track what we add so it can be removed later.
            # (This is for catching return or yield in the finally, which is usually
            # very bad)
            if not any(
                h.type is None
                or (isinstance(h.type, ast.Name) and h.type.id == "BaseException")
                for h in node.handlers
            ):
                self.uncheckpointed_statements.update(added)

            self.visit_nodes(node.finalbody)
            self.uncheckpointed_statements.difference_update(added)

    # valid checkpoint if both body and orelse checkpoint
    def visit_If(self, node: ast.If | ast.IfExp):
        if not self.async_function:
            return
        # visit condition
        self.visit_nodes(node.test)
        outer = self.uncheckpointed_statements.copy()

        # visit body
        self.visit_nodes(node.body)
        body_outer = self.uncheckpointed_statements

        # reset to after condition and visit orelse
        self.uncheckpointed_statements = outer
        self.visit_nodes(node.orelse)

        # union of both branches is the new set of unhandled entries
        self.uncheckpointed_statements.update(body_outer)

    # inline if
    visit_IfExp = visit_If

    # Check for yields w/o checkpoint in between due to entering loop body the first time,
    # after completing all of loop body, and after any continues.
    # yield in else have same requirement
    # state after the loop same as above, and in addition the state at any break
    def visit_loop(self, node: ast.While | ast.For | ast.AsyncFor):
        if not self.async_function:
            return
        # visit condition

        # the following block is duplicated in Visitor103_104
        infinite_loop = False
        if isinstance(node, ast.While):
            try:
                infinite_loop = body_guaranteed_once = bool(ast.literal_eval(node.test))
            except Exception:  # noqa: PIE786
                body_guaranteed_once = False
            self.visit_nodes(node.test)
        else:
            self.visit_nodes(node.target)
            self.visit_nodes(node.iter)
            body_guaranteed_once = iter_guaranteed_once(node.iter)

        # save state in case of nested loops
        outer = self.get_state(
            "uncheckpointed_before_continue",
            "uncheckpointed_before_break",
            "suppress_errors",
        )

        self.uncheckpointed_before_continue = set()

        # AsyncFor guaranteed checkpoint at every iteration
        if isinstance(node, ast.AsyncFor):
            self.uncheckpointed_statements = set()

        pre_body_uncheckpointed_statements = self.uncheckpointed_statements

        # check for all possible uncheckpointed statements before entering start of loop
        # due to `continue` or multiple iterations
        if not isinstance(node, ast.AsyncFor):
            # reset uncheckpointed_statements to not clear it if awaited
            self.uncheckpointed_statements = set()

            # avoid duplicate errors
            self.suppress_errors = True

            # set self.uncheckpointed_before_continue and uncheckpointed at end of loop
            self.visit_nodes(node.body)

            self.suppress_errors = outer["suppress_errors"]

            self.uncheckpointed_statements.update(self.uncheckpointed_before_continue)

        # add uncheckpointed on first iter
        self.uncheckpointed_statements.update(pre_body_uncheckpointed_statements)

        # visit body
        self.uncheckpointed_before_break = set()
        self.visit_nodes(node.body)

        # AsyncFor guarantees checkpoint on running out of iterable
        # so reset checkpoint state at end of loop. (but not state at break)
        if isinstance(node, ast.AsyncFor):
            self.uncheckpointed_statements = set()
        else:
            # enter orelse with worst case:
            # loop body might execute fully before entering orelse
            # (current state of self.uncheckpointed_statements)
            # or not at all
            if not body_guaranteed_once:
                self.uncheckpointed_statements.update(
                    pre_body_uncheckpointed_statements
                )
            # or at a continue, unless it's an infinite loop
            if not infinite_loop:
                self.uncheckpointed_statements.update(
                    self.uncheckpointed_before_continue
                )

        # visit orelse
        self.visit_nodes(node.orelse)

        # if this is an infinite loop, with no break in it, don't raise
        # alarms about the state after it.
        if infinite_loop and not any(
            isinstance(n, ast.Break) for n in self.walk(*node.body)
        ):
            self.uncheckpointed_statements = set()
        else:
            # We may exit from:
            # orelse (covering: no body, body until continue, and all body)
            # break
            self.uncheckpointed_statements.update(self.uncheckpointed_before_break)

        # reset break & continue in case of nested loops
        self.set_state(outer)

    visit_While = visit_loop
    visit_For = visit_loop
    visit_AsyncFor = visit_loop

    # save state in case of continue/break at a point not guaranteed to checkpoint
    def visit_Continue(self, node: ast.Continue):
        if not self.async_function:
            return
        self.uncheckpointed_before_continue.update(self.uncheckpointed_statements)

    def visit_Break(self, node: ast.Break):
        if not self.async_function:
            return
        self.uncheckpointed_before_break.update(self.uncheckpointed_statements)

    # first node in a condition is always evaluated, but may shortcut at any point
    # after that so we track worst-case checkpoint (i.e. after yield)
    def visit_BoolOp(self, node: ast.BoolOp):
        if not self.async_function:
            return
        self.visit(node.op)

        # first value always evaluated
        self.visit(node.values[0])

        worst_case_shortcut = self.uncheckpointed_statements.copy()

        for value in node.values[1:]:
            self.visit(value)
            worst_case_shortcut.update(self.uncheckpointed_statements)

        self.uncheckpointed_statements = worst_case_shortcut



# type: ignore
# ARG --enable-visitor-codes-regex=(TRIO220)
# NOTRIO

# anyio eval will automatically prepend this test with `--anyio`
import trio  # isort: skip


async def foo():
    subprocess.Popen()  # TRIO220: 4, 'subprocess.Popen', "[anyio|trio]"



# NOANYIO
from typing import Any

import trio

par: Any = ...


async def foo():
    # not async
    trio.run(par)

    # safely awaited
    await trio.aclose_forcefully(par)
    await trio.open_file(par)
    await trio.open_ssl_over_tcp_listeners(par, par)
    await trio.open_ssl_over_tcp_stream(par, par)
    await trio.open_tcp_listeners(par)
    await trio.open_tcp_stream(par, par)
    await trio.open_unix_socket(par)
    await trio.run_process(par)
    await trio.serve_listeners(par, par)
    await trio.serve_ssl_over_tcp()
    await trio.serve_tcp()
    await trio.sleep()
    await trio.sleep_forever()
    await trio.sleep_until()
    await trio.lowlevel.cancel_shielded_checkpoint()
    await trio.lowlevel.checkpoint()
    await trio.lowlevel.checkpoint_if_cancelled()
    await trio.lowlevel.open_process()
    await trio.lowlevel.permanently_detach_coroutine_object()
    await trio.lowlevel.reattach_detached_coroutine_object()
    await trio.lowlevel.temporarily_detach_coroutine_object()
    await trio.lowlevel.wait_readable()
    await trio.lowlevel.wait_task_rescheduled()
    await trio.lowlevel.wait_writable()

    # all async functions
    trio.aclose_forcefully()  # error: 4, "trio.aclose_forcefully", "function"
    trio.open_file()  # error: 4, "trio.open_file", "function"
    trio.open_ssl_over_tcp_listeners()  # error: 4, "trio.open_ssl_over_tcp_listeners", "function"
    trio.open_ssl_over_tcp_stream()  # error: 4, "trio.open_ssl_over_tcp_stream", "function"
    trio.open_tcp_listeners()  # error: 4, "trio.open_tcp_listeners", "function"
    trio.open_tcp_stream()  # error: 4, "trio.open_tcp_stream", "function"
    trio.open_unix_socket()  # error: 4, "trio.open_unix_socket", "function"
    trio.run_process()  # error: 4, "trio.run_process", "function"
    trio.serve_listeners()  # error: 4, "trio.serve_listeners", "function"
    trio.serve_ssl_over_tcp()  # error: 4, "trio.serve_ssl_over_tcp", "function"
    trio.serve_tcp()  # error: 4, "trio.serve_tcp", "function"
    trio.sleep()  # error: 4, "trio.sleep", "function"
    trio.sleep_forever()  # error: 4, "trio.sleep_forever", "function"
    trio.sleep_until()  # error: 4, "trio.sleep_until", "function"

    # long lines, wheee
    trio.lowlevel.cancel_shielded_checkpoint()  # error: 4, "trio.lowlevel.cancel_shielded_checkpoint", "function"
    trio.lowlevel.checkpoint()  # error: 4, "trio.lowlevel.checkpoint", "function"
    trio.lowlevel.checkpoint_if_cancelled()  # error: 4, "trio.lowlevel.checkpoint_if_cancelled", "function"
    trio.lowlevel.open_process()  # error: 4, "trio.lowlevel.open_process", "function"
    trio.lowlevel.permanently_detach_coroutine_object()  # error: 4, "trio.lowlevel.permanently_detach_coroutine_object", "function"
    trio.lowlevel.reattach_detached_coroutine_object()  # error: 4, "trio.lowlevel.reattach_detached_coroutine_object", "function"
    trio.lowlevel.temporarily_detach_coroutine_object()  # error: 4, "trio.lowlevel.temporarily_detach_coroutine_object", "function"
    trio.lowlevel.wait_readable()  # error: 4, "trio.lowlevel.wait_readable", "function"
    trio.lowlevel.wait_task_rescheduled()  # error: 4, "trio.lowlevel.wait_task_rescheduled", "function"
    trio.lowlevel.wait_writable()  # error: 4, "trio.lowlevel.wait_writable", "function"

    # safe
    async with await trio.open_file() as f:
        pass

    async with trio.open_file() as f:  # error: 15, "trio.open_file", "function"
        pass

    # safe in theory, but deemed sufficiently poor style that parsing
    # it isn't supported
    k = trio.open_file()  # error: 8, "trio.open_file", "function"
    await k

    # issue #56
    nursery = trio.open_nursery()
    await nursery.start()
    await nursery.start_foo()

    nursery.start()  # error: 4, "trio.Nursery.start", "method"
    None.start()  # type: ignore
    nursery.start_soon()
    nursery.start_foo()

    with trio.open_nursery() as booboo:
        booboo.start()  # error: 8, "trio.Nursery.start", "method"

    barbar: trio.Nursery = ...
    barbar.start()  # error: 4, "trio.Nursery.start", "method"



from typing import Any

import anyio
from anyio import get_cancelled_exc_class, get_cancelled_exc_class as foo  # TRIO118: 0

bar1 = anyio.get_cancelled_exc_class  # TRIO118: 7
bar2 = anyio.get_cancelled_exc_class()  # TRIO118: 7
bar3 = get_cancelled_exc_class  # TRIO118: 7
bar4 = get_cancelled_exc_class()  # TRIO118: 7

bar5: Any = anyio.get_cancelled_exc_class  # TRIO118: 12
bar6: Any = anyio.get_cancelled_exc_class()  # TRIO118: 12
bar7: Any = get_cancelled_exc_class  # TRIO118: 12
bar8: Any = get_cancelled_exc_class()  # TRIO118: 12

# code coverage
bar9: Any
from anyio import sleep  # isort: skip



"""Additional tests for command line parameters and decorator handling."""

from __future__ import annotations

import ast
from pathlib import Path

from flake8.main.application import Application

from flake8_trio.base import Statement
from flake8_trio.visitors.helpers import fnmatch_qualified_name
from flake8_trio.visitors.visitor91x import Visitor91X


def dec_list(*decorators: str) -> ast.Module:
    source = ""
    for dec in decorators:
        source += f"@{dec}\n"
    source += "async def f():\n  bar()"
    return ast.parse(source)


def wrap(decorators: tuple[str, ...], decs2: str) -> str | None:
    tree = dec_list(*decorators)
    assert isinstance(tree.body[0], ast.AsyncFunctionDef)
    return fnmatch_qualified_name(tree.body[0].decorator_list, decs2)


def test_basic():
    assert wrap(("foo",), "foo")
    assert wrap(("foo", "bar"), "foo")
    assert wrap(("bar", "foo"), "foo")

    assert not wrap(("foo",), "foob")


def test_dotted():
    assert wrap(("foo.bar",), "foo.bar")

    assert not wrap(("foo.bar",), "foo")
    assert not wrap(("foo.bar",), "bar")

    assert not wrap(("foo",), "foo.bar")
    assert not wrap(("bar",), "foo.bar")

    assert not wrap(("foo.bar.jane",), "foo.bar")
    assert not wrap(("foo.bar",), "foo.bar.jane")
    assert not wrap(("jane.foo.bar",), "foo.bar")
    assert not wrap(("foo.bar",), "jane.foo.bar")


def test_multidotted():
    assert wrap(("foo.bar.jane",), "foo.bar.jane")
    assert not wrap(("foo.bar",), "foo.bar.jane")
    assert not wrap(("foo.bar.jane",), "foo.bar")


def test_wildcard():
    assert wrap(("foo",), "*")
    assert wrap(("foo.bar",), "*")
    assert not wrap(("foo",), "foo.*")
    assert wrap(("foo.bar",), "foo.*")
    assert not wrap(("bar.foo",), "foo.*")

    assert wrap(("foo",), "foo*")
    assert wrap(("foobar",), "foo*")
    assert wrap(("foobar.bar",), "foo*")
    assert wrap(("foo.bar",), "*.bar")


def test_at():
    assert wrap(("foo",), "@foo")
    assert wrap(("foo.bar",), "@foo.bar")


def test_calls():
    assert wrap(("foo()",), "@foo")
    assert wrap(("foo(1, 2, *x, **y)",), "@foo")
    assert wrap(("foo.bar()",), "@foo.bar")
    assert wrap(("foo.bar(1, 2, *x, **y)",), "@foo.bar")


def test_pep614():
    # Just don't crash and we'll be good.
    assert not wrap(("(any, expression, we, like)",), "no match here")


file_path = str(Path(__file__).parent / "trio_options.py")
common_flags = ["--select=TRIO", file_path]


def test_command_line_1(capfd):
    Application().run(common_flags + ["--no-checkpoint-warning-decorators=app.route"])
    assert capfd.readouterr() == ("", "")


expected_lineno = -1
with open(file_path) as f:
    for lineno, line in enumerate(f, start=1):
        if line.startswith("async def"):
            expected_lineno = lineno
            break

expected_out = (
    f"{file_path}:{expected_lineno}:1: TRIO910 "
    + Visitor91X.error_codes["TRIO910"].format(
        "exit", Statement("function definition", expected_lineno)
    )
    + "\n"
)


def test_command_line_2(capfd):
    Application().run(common_flags + ["--no-checkpoint-warning-decorators=app"])
    assert capfd.readouterr() == (expected_out, "")


def test_command_line_3(capfd):
    Application().run(common_flags)
    assert capfd.readouterr() == (expected_out, "")



# type: ignore
# ANYIO_NO_ERROR
# Conventional usage
from trio import MultiError

try:
    raise MultiError  # TRIO117: 10, "MultiError"
    raise NonBaseMultiError  # TRIO117: 10, "NonBaseMultiError"
    raise trio.MultiError  # TRIO117: 10, "trio.MultiError"
    raise trio.NonBaseMultiError  # TRIO117: 10, "trio.NonBaseMultiError"
except MultiError:  # TRIO117: 7, "MultiError"
    ...
except NonBaseMultiError:  # TRIO117: 7, "NonBaseMultiError"
    ...
except trio.MultiError:  # TRIO117: 7, "trio.MultiError"
    ...
except trio.NonBaseMultiError:  # TRIO117: 7, "trio.NonBaseMultiError"
    ...

# Any other mention of it
MultiError  # TRIO117: 0, "MultiError"
NonBaseMultiError  # TRIO117: 0, "NonBaseMultiError"
trio.MultiError  # TRIO117: 0, "trio.MultiError"
trio.NonBaseMultiError  # TRIO117: 0, "trio.NonBaseMultiError"

MultiError.foo  # TRIO117: 0, "MultiError"
trio.MultiError.foo  # TRIO117: 0, "trio.MultiError"

MultiError()  # TRIO117: 0, "MultiError"


def bar(x: MultiError):  # TRIO117: 11, "MultiError"
    ...


# Known false alarm
MultiError: int  # TRIO117: 0, "MultiError"


# args are not ast.Name's, so this one (surprisingly!) isn't a false positive
# (though any use of the variable will be)
def foo(MultiError: int):
    ...


# only triggers on *trio*.MultiError
anything.MultiError
blah.MultiError.bee



"""Test formatting in error codes."""

from flake8_trio.visitors import ERROR_CLASSES


def test_error_messages_ends_with_period():
    for error_class in ERROR_CLASSES:
        for code, message in error_class.error_codes.items():
            assert message.endswith("."), f"{code} in {error_class} missing period."



"""TRIO105: library async function must be immediately awaited."""

from __future__ import annotations

import ast

from .flake8triovisitor import Flake8TrioVisitor
from .helpers import error_class

# used in 105
trio_async_funcs = (
    "trio.aclose_forcefully",
    "trio.open_file",
    "trio.open_ssl_over_tcp_listeners",
    "trio.open_ssl_over_tcp_stream",
    "trio.open_tcp_listeners",
    "trio.open_tcp_stream",
    "trio.open_unix_socket",
    "trio.run_process",
    "trio.serve_listeners",
    "trio.serve_ssl_over_tcp",
    "trio.serve_tcp",
    "trio.sleep",
    "trio.sleep_forever",
    "trio.sleep_until",
    "trio.lowlevel.cancel_shielded_checkpoint",
    "trio.lowlevel.checkpoint",
    "trio.lowlevel.checkpoint_if_cancelled",
    "trio.lowlevel.open_process",
    "trio.lowlevel.permanently_detach_coroutine_object",
    "trio.lowlevel.reattach_detached_coroutine_object",
    "trio.lowlevel.temporarily_detach_coroutine_object",
    "trio.lowlevel.wait_readable",
    "trio.lowlevel.wait_task_rescheduled",
    "trio.lowlevel.wait_writable",
)


@error_class
class Visitor105(Flake8TrioVisitor):
    error_codes = {
        "TRIO105": "{0} async {1} must be immediately awaited.",
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.typed_calls["trio.open_nursery"] = "trio.Nursery"

    def visit_Call(self, node: ast.Call):
        if getattr(node, "awaited", False) or "trio" not in self.library:
            return

        if (name := ast.unparse(node.func)) in trio_async_funcs:
            self.error(node, name, "function")
        elif isinstance(node.func, ast.Attribute) and node.func.attr == "start":
            var = ast.unparse(node.func.value)

            if self.variables.get(var, "") == "trio.Nursery" or var.endswith("nursery"):
                self.error(node, "trio.Nursery.start", "method")
